/******************************************************************************/

/* mrtab : Miller-Rabin test iterations for a k-bit, random probable prime
 * search, satisfying an upper bound for the error-probability: p(k, t)
 *
 * Copyright (C) 2006, 2007 Brett Hale.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

/* constructive feedback, improvements, suggestions and bug reports are
 * welcome. email: brettyhale@yah00.c0m.au */


/* [1] Damgaard, Landrock, Pomerance, "Average Case Error Estimates for
 * the Strong Probable Prime Test," Mathematics of Computation, Vol. 61,
 * Jul. 1993, pp 177 - 194. aka: [DLP] (referenced in HAC.4.4.1). */

/* M-R iterations for (k > 16) are computed, so that trial division can
 * be performed with 16-bit arithmetic. the probablility that a random,
 * k-bit odd(N) is prime, according to DLP.proposition.2 satisfies:

 * (pi(2^k) - pi(2^(k - 1))) / 2^(k - 2) > (0.71867 * 4) / k ; k >= 21.

 * by counting the primes < (2^20), it is clear that this lower bound is
 * actually valid for: k >= 8. */


/* [2] R. Burthe, Jr., "Further Investigations with the Strong Probable
 * Prime Test," Mathematics of Computation, Vol. 65, Jan. 1996, pp 373 -
 * 381. aka: [RBJ] */

/* an important result in [2] is that it validates the conjecture in [1]
 * that: p(k, t) <= (4^-t), for all k >= 2, t >= 1. */


/* [HAC] Menezes, Oorschot, Vanstone, "Handbook of Applied Cryptography,"
 * CRC Press, ISBN 0849385237. also available online (conditionally) at:
 * [http://www.cacr.math.uwaterloo.ca/hac/] */

/* chapter 4, in particular section 4.4.1, introduces the random probable
 * prime search using the Miller-Rabin test. */

/******************************************************************************/

#include <stddef.h>
#include <string.h>

#include <float.h>
#include <math.h>

#ifndef M_PI
#define M_PI (3.14159265358979323846)
#endif

#ifndef M_LN2
#define M_LN2 (0.69314718055994530942)
#endif

#include <stdio.h>


/* return (1) if the nul-terminated C string forms a valid
 * 32-bit unsigned integer value in C locale decimal format;
 * return (0) otherwise: */

static int u32_arg (unsigned long *u, const char *s)
{
    unsigned long x = 0, d;

    for (; (d = (unsigned long) (*s++)) != 0; x += d)
    {
        if ((d -= (unsigned long) ('0')) > (9) ||
            (x > (0xffffffffUL / (10)))) return (0);
        if ((x *= (10)) > (0xffffffffUL - d))
            return (0);
    }

    *u = x; return (1); /* 32-bit unsigned integer value. */
}

/******************************************************************************/

/* the signature for the p(k, t) evaluation function: */

/* note: an implementation may assert that: (k > 16 && t >= 1), although
 * it should attempt to handle all (k > 1). the values of (k <= 1) yield:
 * p(k, t) = (1) (no primes), while (t = 0) is meaningless. */

typedef double (*p_kt_fn)(unsigned int, unsigned int);

static int dlp_tab (p_kt_fn p_kt)
{
    unsigned int k, t;

    fprintf(stdout, "lower bounds for -lg(p(k, t))\n\n");

    fprintf(stdout, "k\\t |");
    for (t = 1; t <= 10; t++)
        fprintf(stdout, " %3u", t);
    fprintf(stdout, "\n-----");
    for (t = 1; t <= 10; t++)
        fprintf(stdout, "----");
    fprintf(stdout, "\n");

    for (k = 100; k <= 600; k += 50)
    {
        fprintf(stdout, "%3u |", k);
        for (t = 1; t <= 10; t++)
            fprintf(stdout, " %3u",
                    (unsigned int) (- log((*p_kt)(k, t)) / M_LN2));
        fprintf(stdout, "\n");
    }

    return (0);
}


/* RBJ.4 : exact p(k, 1) values for 2 <= k <= 24 (Monier's result) : */

/* the finite precision values in this table are never less than p(k, 1).
 * since p(k, 1) < 1/5 for 2 <= k <= 24, the Monier-Rabin theorem yields:
 * p(k, t) <= 4^(1-t) * p(k, 1) / (1 - p(k, 1)) < (4^-t). */

static const double p_k1_lut[25] = /* 'exact' p(0, 1) .. p(24, 1) */
{
    1.000000000000000e+00, 1.000000000000000e+00, 0.000000000000000e+00,
    0.000000000000000e+00, 1.641791044776120e-01, 6.429942418426104e-02,
    6.534806483607850e-02, 5.665475225100304e-02, 3.800377817839186e-02,
    3.083711963540037e-02, 2.052507976426565e-02, 1.739357468061933e-02,
    1.071035918278328e-02, 7.949087169865036e-03, 5.933704380893274e-03,
    3.944264306920955e-03, 2.625516611747675e-03, 1.928651879061135e-03,
    1.257789417491376e-03, 9.045714725091588e-04, 6.088531201662905e-04,
    4.017062956817300e-04, 2.757637921694800e-04, 1.876065468255186e-04,
    1.261284736535026e-04
};

/******************************************************************************/

/* dlp_kt: this is an implementation of the estimate in DLP.4, with a few
 * simple optimizations. the results match those in DLP.table.1. */

static double dlp_kt (unsigned int k, unsigned int t)
{
    const double c = (8.0 * (M_PI * M_PI - 6.0) / 3.0);

    double rk = (double) k, rt = (double) t, mt, rp;
    unsigned int mh, mi;

    /* assert(k > 1 && t >= 1); */

    rp = pow(4.0, - rt);

    if (k < 25) /* RBJ.4 (Monier-Rabin) : */
    {
        double p_k1 = p_k1_lut[k];

        if (t > 1)
            rp *= (4.0 * p_k1) / (1.0 - p_k1);
        else
            rp = p_k1;

        if (k < 8) return rp; /* no DLP.4 result. */
    }

    rk -= 1.0, mt = pow(2.0, (1.0 - rt));

    /* integral 'M' candidates: */

    mh = (unsigned int) (2.0 * sqrt(rk) - 1.0);
    for (mi = 3; mi <= mh; mi++)
    {
        double r0 = 0.0;
        unsigned int j;

        for (j = 2; j <= mi; j++) /* {j .. M} */
        {
            double ej = (double) j, mj;
            unsigned int m;

            ej += rk / ej, m = j + (j == 2);
            mj = pow(2.0, (((1.0 - rt) * (double) m) - ej));

            for (r0 += mj; m < mi; m++) /* {m .. M} */
                r0 += (mj *= mt);
        }

        r0 *= c / (2.0 * mt);
        r0 += pow(2.0, - (((double) mi) * rt + 2.0));

        if ((r0 *= (rk + 1.0) / 0.71867) < rp) /* new 'M' */
            rp = r0;
    }

    return rp; /* p(k, t). */
}

/* TODO: both DLP.7 and RBJ.5 describe combined results from:

 * [3] S.H. Kim and C. Pomerance, "The Probability that a Random Probable
 * Prime is Composite," Mathematics of Computation, Vol. 53, Oct. 1989,
 * pp 721-742.

 * for 'large' (k/t) ratios, this appears to give much stronger results.
 * the complexity of the estimate is daunting - finding the optimal value
 * is probably not feasible. */

/******************************************************************************/

/* how to use the M-R table: the first entry is the maximum (k) value for
 * which (2) iterations of the M-R test are required to guarantee that:
 * p(k, t) <= (2^-s). an entry with a LUT index of (i) is the maximum (k)
 * value for which (i + 2) iterations are required. */

/* the (0) entry is an 'end-of-table' marker which delimits the maximum
 * number of iterations required for (k > 16). that is, if the (0) entry
 * has a LUT index of (j), then the maximum number of iterations required
 * is (tmax = j + 1). usage: "for (t = 1; k <= lut[t - 1]; t++);" */


static const char *usage[] =
{
    "usage: mrtab [s], where s = 64 .. 256 (default: 128)\n",
    "M-R test iterations s.t. p(k, t) <= (2^-s), for k > 16.\n", NULL
};

int main (int argc, char **argv)
{
    p_kt_fn p_kt = dlp_kt; /* default p(k, t) evaluation function. */

    unsigned int s = (128), tmax, kmax, t, k, ttab[256];
    double pmax;

    if (argc > 1) /* exponent option: */
    {
        unsigned long u;

        if (strcmp(argv[1], "-d") == 0)
            return dlp_tab(p_kt);

        if (!u32_arg(& u, argv[1]) || (u < 64) || (u > 256))
        {
            const char **msg;
            for (msg = usage; *msg != NULL; msg++)
                fprintf(stderr, "%s", *msg);
            return (1);
        }

        s = (unsigned int) u;
    }

    /* safe bias for the error-probability: */
    pmax = pow(2.0 * (1.0 + DBL_EPSILON), - ((double) s));

    fprintf(stdout, "k from t = 2 (k > 16) s.t. ");
    fprintf(stdout, "p(k, t) <= 2^-%u (%.6e) :\n", s, pmax);


    /* find tmax s.t. p(k, t) <= (2^-s), for k > 16. the result from
     * [RBJ] yields: tmax = ceil(s/2), for k >= 2. (s) is an integral
     * value in this context: */

    tmax = (s + 1) >> 1;

    /* under the assertion that: p(k + 1, t) < p(k, t), on which the
     * implementation of a threshold table is predicated, this is the
     * upper bound for t s.t. p(k, t) <= (2^-s), for k > 16. */

    /* find kmax s.t. p(kmax, 1) <= (2^-s). kmax must be greater than,
     * or equal to, the minimum k value s.t. p(k, 1) <= (2^-s) : */

    for (kmax = 17; (*p_kt)(kmax, 1) > pmax; kmax <<= 1);

    /* fprintf(stdout, "tmax = %3u, kmax = %5u\n\n", tmax, kmax); */


    for (k = kmax, t = 2; t <= tmax; t++)
    {
        unsigned int k0 = (17), k1 = k;
        int found = 0;

        /* if: p(k + 1, t - 1) <= pmax <= p(k, t - 1), assert that:
         * p(k, t) < p(k, t - 1). it does not follow (although it is
         * almost certainly the case) that: p(k, t) <= pmax. */

        while (!found && k0 <= k1)
        {
            unsigned int kt = (k0 + k1) >> 1;

            if ((*p_kt)(kt + 1, t - 1) > pmax) /* k > kt */
                k0 = kt + 1;

            else if ((*p_kt)(kt, t - 1) < pmax) /* k < kt */
                k1 = kt - 1;

            else /* sweet-spot: */
            {
                while ((*p_kt)(kt, t) > pmax) kt++; /* (check) */

                if (kt > k) /* pathological case (?) */
                {
                    unsigned int ti;

                    /* warning: there is a local maxima occuring within
                     * p(k, t). ensure that the table contains a sequence
                     * of non-increasing (k) values: */

                    for (ti = t - 1; ti >= 2 && ttab[ti] < kt; ti--)
                        ttab[ti] = kt;
                }

                found = 1, ttab[t] = (k = kt);
                /* fprintf(stdout, "t = %3u @ k = %5u\n", t, k); */
            }
        }

        /* if no threshold (k) value was found, then tmax has been
         * determined to be unnecessarily high by p(k, t). clamp the
         * table to its current size: */

        if (!found) tmax = t - 1;
    }


    /* print threshold values in a LUT format: */

    fprintf(stdout, "\n");
    fprintf(stdout, "    0x%04x", ttab[2]);

    ttab[++tmax] = 0; /* EOT marker. */

    for (t = 3; t <= tmax; t++)
    {
        if ((t - 2) % 8 != 0)
            fprintf(stdout, ", 0x%04x", ttab[t]);
        else
            fprintf(stdout, ",\n    0x%04x", ttab[t]);
    }

    fprintf(stdout, "\n\n");

    /* the M-R implementation must handle candidates with less than (17)
     * significant bits explicitly. this requires a maximum of (54) trial
     * divisions, i.e., the number of primes less than (2^8). */

    return (0);
}

/******************************************************************************/
